from typing import Literal
from .hero import Hero, HeroSet
from enum import Enum
from ..exceptions import NumberOfPlayersError, HeroRepeatError


DEFAULT_NUMBER_OF_PLAYERS = 5


class Side(str, Enum):
    RADIANT = "radiant"
    DIRE = "dire"


class Team(object):
    def __init__(self, 
                 heroes: list[Hero], 
                 *,
                 side: Literal[Side.RADIANT, Side.DIRE]=None,
                 name: str=None, 
                 is_winner: bool=None,
                 no_repeat: bool=True,
                 five_players: bool=True) -> None:
        self.heroes = HeroSet(heroes)
        self.side = side
        self.name = name
        self.is_winner = is_winner
        if five_players and len(heroes) != DEFAULT_NUMBER_OF_PLAYERS:
            raise NumberOfPlayersError()
        if no_repeat and len(set(heroes)) != len(heroes):
            raise HeroRepeatError()


class Radiant(Team):
    def __init__(self, 
                 heroes: list[Hero], 
                 *,
                 name: str = None, 
                 is_winner: bool = None, 
                 no_repeat: bool=True,
                 five_players: bool=True) -> None:
        super().__init__(heroes, 
                         side=Side.RADIANT, 
                         name=name, 
                         is_winner=is_winner, 
                         no_repeat=no_repeat,
                         five_players=five_players)
    

class Dire(Team):
    def __init__(self, 
                 heroes: list[Hero], 
                 *,
                 name: str = None, 
                 is_winner: bool = None, 
                 no_repeat: bool=True,
                 five_players: bool=True) -> None:
        super().__init__(heroes, 
                         side=Side.DIRE, 
                         name=name, 
                         is_winner=is_winner, 
                         no_repeat=no_repeat,
                         five_players=five_players)


from dataclasses import dataclass
from functools import cached_property
from ..data.hero import all_heroes_json


@dataclass
class Hero(object):
    id: int
    sysname: str
    fullname: str
    name: str

    def __hash__(self) -> int:
        return self.id


_all_heroes_map = {}
_all_heroes_list = []
for hero_json in all_heroes_json:
    hero = Hero(**hero_json)
    _all_heroes_list.append(hero)
    for attr in vars(hero).values():
        _all_heroes_map[attr] = hero
del hero_json
del attr


class HeroSet(object):
    def __init__(self, heroes: list[Hero]) -> None:
        self._heroes = heroes

    def get_heroes_fields(self, field_name: str):
        return [getattr(hero, field_name) for hero in self._heroes]

    def find(self, value):
        return _all_heroes_map.get(value, None)

    @property
    def heroes(self):
        return self._heroes
    
    @cached_property
    def ids(self):
        return self.get_heroes_fields("id")
    
    @cached_property
    def sysnames(self):
        return self.get_heroes_fields("sysname")
    
    @cached_property
    def fullnames(self):
        return self.get_heroes_fields("fullname")
    
    @cached_property
    def names(self):
        return self.get_heroes_fields("name")
    
    def __iter__(self):
        return iter(self._heroes)

    def __add__(self, heroset: "HeroSet"):
        heroes = self._heroes + heroset.heroes
        return HeroSet(heroes)


all_heroes = HeroSet(_all_heroes_list)
find_hero = all_heroes.find


all_heroes_json = [
    {
        "id": 1,
        "sysname": "npc_dota_hero_antimage",
        "fullname": "Anti-Mage",
        "name": "antimage"
    },
    {
        "id": 2,
        "sysname": "npc_dota_hero_axe",
        "fullname": "Axe",
        "name": "axe"
    },
    {
        "id": 3,
        "sysname": "npc_dota_hero_bane",
        "fullname": "Bane",
        "name": "bane"
    },
    {
        "id": 4,
        "sysname": "npc_dota_hero_bloodseeker",
        "fullname": "Bloodseeker",
        "name": "bloodseeker"
    },
    {
        "id": 5,
        "sysname": "npc_dota_hero_crystal_maiden",
        "fullname": "Crystal Maiden",
        "name": "crystalmaiden"
    },
    {
        "id": 6,
        "sysname": "npc_dota_hero_drow_ranger",
        "fullname": "Drow Ranger",
        "name": "drowranger"
    },
    {
        "id": 7,
        "sysname": "npc_dota_hero_earthshaker",
        "fullname": "Earthshaker",
        "name": "earthshaker"
    },
    {
        "id": 8,
        "sysname": "npc_dota_hero_juggernaut",
        "fullname": "Juggernaut",
        "name": "juggernaut"
    },
    {
        "id": 9,
        "sysname": "npc_dota_hero_mirana",
        "fullname": "Mirana",
        "name": "mirana"
    },
    {
        "id": 10,
        "sysname": "npc_dota_hero_morphling",
        "fullname": "Morphling",
        "name": "morphling"
    },
    {
        "id": 11,
        "sysname": "npc_dota_hero_nevermore",
        "fullname": "Shadow Fiend",
        "name": "shadowfiend"
    },
    {
        "id": 12,
        "sysname": "npc_dota_hero_phantom_lancer",
        "fullname": "Phantom Lancer",
        "name": "phantomlancer"
    },
    {
        "id": 13,
        "sysname": "npc_dota_hero_puck",
        "fullname": "Puck",
        "name": "puck"
    },
    {
        "id": 14,
        "sysname": "npc_dota_hero_pudge",
        "fullname": "Pudge",
        "name": "pudge"
    },
    {
        "id": 15,
        "sysname": "npc_dota_hero_razor",
        "fullname": "Razor",
        "name": "razor"
    },
    {
        "id": 16,
        "sysname": "npc_dota_hero_sand_king",
        "fullname": "Sand King",
        "name": "sandking"
    },
    {
        "id": 17,
        "sysname": "npc_dota_hero_storm_spirit",
        "fullname": "Storm Spirit",
        "name": "stormspirit"
    },
    {
        "id": 18,
        "sysname": "npc_dota_hero_sven",
        "fullname": "Sven",
        "name": "sven"
    },
    {
        "id": 19,
        "sysname": "npc_dota_hero_tiny",
        "fullname": "Tiny",
        "name": "tiny"
    },
    {
        "id": 20,
        "sysname": "npc_dota_hero_vengefulspirit",
        "fullname": "Vengeful Spirit",
        "name": "vengefulspirit"
    },
    {
        "id": 21,
        "sysname": "npc_dota_hero_windrunner",
        "fullname": "Windranger",
        "name": "windranger"
    },
    {
        "id": 22,
        "sysname": "npc_dota_hero_zuus",
        "fullname": "Zeus",
        "name": "zeus"
    },
    {
        "id": 23,
        "sysname": "npc_dota_hero_kunkka",
        "fullname": "Kunkka",
        "name": "kunkka"
    },
    {
        "id": 25,
        "sysname": "npc_dota_hero_lina",
        "fullname": "Lina",
        "name": "lina"
    },
    {
        "id": 26,
        "sysname": "npc_dota_hero_lion",
        "fullname": "Lion",
        "name": "lion"
    },
    {
        "id": 27,
        "sysname": "npc_dota_hero_shadow_shaman",
        "fullname": "Shadow Shaman",
        "name": "shadowshaman"
    },
    {
        "id": 28,
        "sysname": "npc_dota_hero_slardar",
        "fullname": "Slardar",
        "name": "slardar"
    },
    {
        "id": 29,
        "sysname": "npc_dota_hero_tidehunter",
        "fullname": "Tidehunter",
        "name": "tidehunter"
    },
    {
        "id": 30,
        "sysname": "npc_dota_hero_witch_doctor",
        "fullname": "Witch Doctor",
        "name": "witchdoctor"
    },
    {
        "id": 31,
        "sysname": "npc_dota_hero_lich",
        "fullname": "Lich",
        "name": "lich"
    },
    {
        "id": 32,
        "sysname": "npc_dota_hero_riki",
        "fullname": "Riki",
        "name": "riki"
    },
    {
        "id": 33,
        "sysname": "npc_dota_hero_enigma",
        "fullname": "Enigma",
        "name": "enigma"
    },
    {
        "id": 34,
        "sysname": "npc_dota_hero_tinker",
        "fullname": "Tinker",
        "name": "tinker"
    },
    {
        "id": 35,
        "sysname": "npc_dota_hero_sniper",
        "fullname": "Sniper",
        "name": "sniper"
    },
    {
        "id": 36,
        "sysname": "npc_dota_hero_necrolyte",
        "fullname": "Necrophos",
        "name": "necrophos"
    },
    {
        "id": 37,
        "sysname": "npc_dota_hero_warlock",
        "fullname": "Warlock",
        "name": "warlock"
    },
    {
        "id": 38,
        "sysname": "npc_dota_hero_beastmaster",
        "fullname": "Beastmaster",
        "name": "beastmaster"
    },
    {
        "id": 39,
        "sysname": "npc_dota_hero_queenofpain",
        "fullname": "Queen of Pain",
        "name": "queenofpain"
    },
    {
        "id": 40,
        "sysname": "npc_dota_hero_venomancer",
        "fullname": "Venomancer",
        "name": "venomancer"
    },
    {
        "id": 41,
        "sysname": "npc_dota_hero_faceless_void",
        "fullname": "Faceless Void",
        "name": "facelessvoid"
    },
    {
        "id": 42,
        "sysname": "npc_dota_hero_skeleton_king",
        "fullname": "Wraith King",
        "name": "wraithking"
    },
    {
        "id": 43,
        "sysname": "npc_dota_hero_death_prophet",
        "fullname": "Death Prophet",
        "name": "deathprophet"
    },
    {
        "id": 44,
        "sysname": "npc_dota_hero_phantom_assassin",
        "fullname": "Phantom Assassin",
        "name": "phantomassassin"
    },
    {
        "id": 45,
        "sysname": "npc_dota_hero_pugna",
        "fullname": "Pugna",
        "name": "pugna"
    },
    {
        "id": 46,
        "sysname": "npc_dota_hero_templar_assassin",
        "fullname": "Templar Assassin",
        "name": "templarassassin"
    },
    {
        "id": 47,
        "sysname": "npc_dota_hero_viper",
        "fullname": "Viper",
        "name": "viper"
    },
    {
        "id": 48,
        "sysname": "npc_dota_hero_luna",
        "fullname": "Luna",
        "name": "luna"
    },
    {
        "id": 49,
        "sysname": "npc_dota_hero_dragon_knight",
        "fullname": "Dragon Knight",
        "name": "dragonknight"
    },
    {
        "id": 50,
        "sysname": "npc_dota_hero_dazzle",
        "fullname": "Dazzle",
        "name": "dazzle"
    },
    {
        "id": 51,
        "sysname": "npc_dota_hero_rattletrap",
        "fullname": "Clockwerk",
        "name": "clockwerk"
    },
    {
        "id": 52,
        "sysname": "npc_dota_hero_leshrac",
        "fullname": "Leshrac",
        "name": "leshrac"
    },
    {
        "id": 53,
        "sysname": "npc_dota_hero_furion",
        "fullname": "Nature's Prophet",
        "name": "naturesprophet"
    },
    {
        "id": 54,
        "sysname": "npc_dota_hero_life_stealer",
        "fullname": "Lifestealer",
        "name": "lifestealer"
    },
    {
        "id": 55,
        "sysname": "npc_dota_hero_dark_seer",
        "fullname": "Dark Seer",
        "name": "darkseer"
    },
    {
        "id": 56,
        "sysname": "npc_dota_hero_clinkz",
        "fullname": "Clinkz",
        "name": "clinkz"
    },
    {
        "id": 57,
        "sysname": "npc_dota_hero_omniknight",
        "fullname": "Omniknight",
        "name": "omniknight"
    },
    {
        "id": 58,
        "sysname": "npc_dota_hero_enchantress",
        "fullname": "Enchantress",
        "name": "enchantress"
    },
    {
        "id": 59,
        "sysname": "npc_dota_hero_huskar",
        "fullname": "Huskar",
        "name": "huskar"
    },
    {
        "id": 60,
        "sysname": "npc_dota_hero_night_stalker",
        "fullname": "Night Stalker",
        "name": "nightstalker"
    },
    {
        "id": 61,
        "sysname": "npc_dota_hero_broodmother",
        "fullname": "Broodmother",
        "name": "broodmother"
    },
    {
        "id": 62,
        "sysname": "npc_dota_hero_bounty_hunter",
        "fullname": "Bounty Hunter",
        "name": "bountyhunter"
    },
    {
        "id": 63,
        "sysname": "npc_dota_hero_weaver",
        "fullname": "Weaver",
        "name": "weaver"
    },
    {
        "id": 64,
        "sysname": "npc_dota_hero_jakiro",
        "fullname": "Jakiro",
        "name": "jakiro"
    },
    {
        "id": 65,
        "sysname": "npc_dota_hero_batrider",
        "fullname": "Batrider",
        "name": "batrider"
    },
    {
        "id": 66,
        "sysname": "npc_dota_hero_chen",
        "fullname": "Chen",
        "name": "chen"
    },
    {
        "id": 67,
        "sysname": "npc_dota_hero_spectre",
        "fullname": "Spectre",
        "name": "spectre"
    },
    {
        "id": 68,
        "sysname": "npc_dota_hero_ancient_apparition",
        "fullname": "Ancient Apparition",
        "name": "ancientapparition"
    },
    {
        "id": 69,
        "sysname": "npc_dota_hero_doom_bringer",
        "fullname": "Doom",
        "name": "doom"
    },
    {
        "id": 70,
        "sysname": "npc_dota_hero_ursa",
        "fullname": "Ursa",
        "name": "ursa"
    },
    {
        "id": 71,
        "sysname": "npc_dota_hero_spirit_breaker",
        "fullname": "Spirit Breaker",
        "name": "spiritbreaker"
    },
    {
        "id": 72,
        "sysname": "npc_dota_hero_gyrocopter",
        "fullname": "Gyrocopter",
        "name": "gyrocopter"
    },
    {
        "id": 73,
        "sysname": "npc_dota_hero_alchemist",
        "fullname": "Alchemist",
        "name": "alchemist"
    },
    {
        "id": 74,
        "sysname": "npc_dota_hero_invoker",
        "fullname": "Invoker",
        "name": "invoker"
    },
    {
        "id": 75,
        "sysname": "npc_dota_hero_silencer",
        "fullname": "Silencer",
        "name": "silencer"
    },
    {
        "id": 76,
        "sysname": "npc_dota_hero_obsidian_destroyer",
        "fullname": "Outworld Destroyer",
        "name": "outworlddestroyer"
    },
    {
        "id": 77,
        "sysname": "npc_dota_hero_lycan",
        "fullname": "Lycan",
        "name": "lycan"
    },
    {
        "id": 78,
        "sysname": "npc_dota_hero_brewmaster",
        "fullname": "Brewmaster",
        "name": "brewmaster"
    },
    {
        "id": 79,
        "sysname": "npc_dota_hero_shadow_demon",
        "fullname": "Shadow Demon",
        "name": "shadowdemon"
    },
    {
        "id": 80,
        "sysname": "npc_dota_hero_lone_druid",
        "fullname": "Lone Druid",
        "name": "lonedruid"
    },
    {
        "id": 81,
        "sysname": "npc_dota_hero_chaos_knight",
        "fullname": "Chaos Knight",
        "name": "chaosknight"
    },
    {
        "id": 82,
        "sysname": "npc_dota_hero_meepo",
        "fullname": "Meepo",
        "name": "meepo"
    },
    {
        "id": 83,
        "sysname": "npc_dota_hero_treant",
        "fullname": "Treant Protector",
        "name": "treantprotector"
    },
    {
        "id": 84,
        "sysname": "npc_dota_hero_ogre_magi",
        "fullname": "Ogre Magi",
        "name": "ogremagi"
    },
    {
        "id": 85,
        "sysname": "npc_dota_hero_undying",
        "fullname": "Undying",
        "name": "undying"
    },
    {
        "id": 86,
        "sysname": "npc_dota_hero_rubick",
        "fullname": "Rubick",
        "name": "rubick"
    },
    {
        "id": 87,
        "sysname": "npc_dota_hero_disruptor",
        "fullname": "Disruptor",
        "name": "disruptor"
    },
    {
        "id": 88,
        "sysname": "npc_dota_hero_nyx_assassin",
        "fullname": "Nyx Assassin",
        "name": "nyxassassin"
    },
    {
        "id": 89,
        "sysname": "npc_dota_hero_naga_siren",
        "fullname": "Naga Siren",
        "name": "nagasiren"
    },
    {
        "id": 90,
        "sysname": "npc_dota_hero_keeper_of_the_light",
        "fullname": "Keeper of the Light",
        "name": "keeperofthelight"
    },
    {
        "id": 91,
        "sysname": "npc_dota_hero_wisp",
        "fullname": "Io",
        "name": "io"
    },
    {
        "id": 92,
        "sysname": "npc_dota_hero_visage",
        "fullname": "Visage",
        "name": "visage"
    },
    {
        "id": 93,
        "sysname": "npc_dota_hero_slark",
        "fullname": "Slark",
        "name": "slark"
    },
    {
        "id": 94,
        "sysname": "npc_dota_hero_medusa",
        "fullname": "Medusa",
        "name": "medusa"
    },
    {
        "id": 95,
        "sysname": "npc_dota_hero_troll_warlord",
        "fullname": "Troll Warlord",
        "name": "trollwarlord"
    },
    {
        "id": 96,
        "sysname": "npc_dota_hero_centaur",
        "fullname": "Centaur Warrunner",
        "name": "centaurwarrunner"
    },
    {
        "id": 97,
        "sysname": "npc_dota_hero_magnataur",
        "fullname": "Magnus",
        "name": "magnus"
    },
    {
        "id": 98,
        "sysname": "npc_dota_hero_shredder",
        "fullname": "Timbersaw",
        "name": "timbersaw"
    },
    {
        "id": 99,
        "sysname": "npc_dota_hero_bristleback",
        "fullname": "Bristleback",
        "name": "bristleback"
    },
    {
        "id": 100,
        "sysname": "npc_dota_hero_tusk",
        "fullname": "Tusk",
        "name": "tusk"
    },
    {
        "id": 101,
        "sysname": "npc_dota_hero_skywrath_mage",
        "fullname": "Skywrath Mage",
        "name": "skywrathmage"
    },
    {
        "id": 102,
        "sysname": "npc_dota_hero_abaddon",
        "fullname": "Abaddon",
        "name": "abaddon"
    },
    {
        "id": 103,
        "sysname": "npc_dota_hero_elder_titan",
        "fullname": "Elder Titan",
        "name": "eldertitan"
    },
    {
        "id": 104,
        "sysname": "npc_dota_hero_legion_commander",
        "fullname": "Legion Commander",
        "name": "legioncommander"
    },
    {
        "id": 105,
        "sysname": "npc_dota_hero_techies",
        "fullname": "Techies",
        "name": "techies"
    },
    {
        "id": 106,
        "sysname": "npc_dota_hero_ember_spirit",
        "fullname": "Ember Spirit",
        "name": "emberspirit"
    },
    {
        "id": 107,
        "sysname": "npc_dota_hero_earth_spirit",
        "fullname": "Earth Spirit",
        "name": "earthspirit"
    },
    {
        "id": 108,
        "sysname": "npc_dota_hero_abyssal_underlord",
        "fullname": "Underlord",
        "name": "underlord"
    },
    {
        "id": 109,
        "sysname": "npc_dota_hero_terrorblade",
        "fullname": "Terrorblade",
        "name": "terrorblade"
    },
    {
        "id": 110,
        "sysname": "npc_dota_hero_phoenix",
        "fullname": "Phoenix",
        "name": "phoenix"
    },
    {
        "id": 111,
        "sysname": "npc_dota_hero_oracle",
        "fullname": "Oracle",
        "name": "oracle"
    },
    {
        "id": 112,
        "sysname": "npc_dota_hero_winter_wyvern",
        "fullname": "Winter Wyvern",
        "name": "winterwyvern"
    },
    {
        "id": 113,
        "sysname": "npc_dota_hero_arc_warden",
        "fullname": "Arc Warden",
        "name": "arcwarden"
    },
    {
        "id": 114,
        "sysname": "npc_dota_hero_monkey_king",
        "fullname": "Monkey King",
        "name": "monkeyking"
    },
    {
        "id": 119,
        "sysname": "npc_dota_hero_dark_willow",
        "fullname": "Dark Willow",
        "name": "darkwillow"
    },
    {
        "id": 120,
        "sysname": "npc_dota_hero_pangolier",
        "fullname": "Pangolier",
        "name": "pangolier"
    },
    {
        "id": 121,
        "sysname": "npc_dota_hero_grimstroke",
        "fullname": "Grimstroke",
        "name": "grimstroke"
    },
    {
        "id": 123,
        "sysname": "npc_dota_hero_hoodwink",
        "fullname": "Hoodwink",
        "name": "hoodwink"
    },
    {
        "id": 126,
        "sysname": "npc_dota_hero_void_spirit",
        "fullname": "Void Spirit",
        "name": "voidspirit"
    },
    {
        "id": 128,
        "sysname": "npc_dota_hero_snapfire",
        "fullname": "Snapfire",
        "name": "snapfire"
    },
    {
        "id": 129,
        "sysname": "npc_dota_hero_mars",
        "fullname": "Mars",
        "name": "mars"
    },
    {
        "id": 135,
        "sysname": "npc_dota_hero_dawnbreaker",
        "fullname": "Dawnbreaker",
        "name": "dawnbreaker"
    },
    {
        "id": 136,
        "sysname": "npc_dota_hero_marci",
        "fullname": "Marci",
        "name": "marci"
    },
    {
        "id": 137,
        "sysname": "npc_dota_hero_primal_beast",
        "fullname": "Primal Beast",
        "name": "primalbeast"
    },
    {
        "id": 138,
        "sysname": "npc_dota_hero_muerta",
        "fullname": "Muerta",
        "name": "muerta"
    }
]


from .team import Team
from functools import cached_property


class Match(object):
    def __init__(self, 
                 radiant: Team, 
                 dire: Team, 
                 *, 
                 map: int=None, 
                 championship: str=None, 
                 id: int=None) -> None:
        self.radiant = radiant
        self.dire = dire
        self.map = map
        self.championship = championship
        self.id = id
    
    @cached_property
    def winner(self):
        if self.radiant.is_winner:
            return self.radiant
        if self.dire.is_winner:
            return self.dire
        return None

    @cached_property
    def loser(self):
        if self.winner is self.radiant:
            return self.dire
        if self.winner is self.dire:
            return self.radiant
        return None
    
    @cached_property
    def heroes(self):
        _heroes = self.radiant.heroes + self.dire.heroes
        return _heroes